# coding=utf-8
"""Rename names in a python module."""
# Copyright 2017 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import ast
import six

from pasta.augment import import_utils
from pasta.base import ast_utils
from pasta.base import scope


def rename_external(t, old_name, new_name):
  """Rename an imported name in a module.

  This will rewrite all import statements in `tree` that reference the old
  module as well as any names in `tree` which reference the imported name. This
  may introduce new import statements, but only if necessary.

  For example, to move and rename the module `foo.bar.utils` to `foo.bar_utils`:
  > rename_external(tree, 'foo.bar.utils', 'foo.bar_utils')

  - import foo.bar.utils
  + import foo.bar_utils

  - from foo.bar import utils
  + from foo import bar_utils

  - from foo.bar import logic, utils
  + from foo.bar import logic
  + from foo import bar_utils

  Arguments:
    t: (ast.Module) Module syntax tree to perform the rename in. This will be
      updated as a result of this function call with all affected nodes changed
      and potentially new Import/ImportFrom nodes added.
    old_name: (string) Fully-qualified path of the name to replace.
    new_name: (string) Fully-qualified path of the name to update to.

  Returns:
    True if any changes were made, False otherwise.
  """
  sc = scope.analyze(t)

  if old_name not in sc.external_references:
    return False

  has_changed = False
  renames = {}
  already_changed = []
  for ref in sc.external_references[old_name]:
    if isinstance(ref.node, ast.alias):
      parent = sc.parent(ref.node)
      # An alias may be the most specific reference to an imported name, but it
      # could if it is a child of an ImportFrom, the ImportFrom node's module
      # may also need to be updated.
      if isinstance(parent, ast.ImportFrom) and parent not in already_changed:
        assert _rename_name_in_importfrom(sc, parent, old_name, new_name)
        renames[old_name.rsplit('.', 1)[-1]] = new_name.rsplit('.', 1)[-1]
        already_changed.append(parent)
      else:
        ref.node.name = new_name + ref.node.name[len(old_name):]
        if not ref.node.asname:
          renames[old_name] = new_name
      has_changed = True
    elif isinstance(ref.node, ast.ImportFrom):
      if ref.node not in already_changed:
        assert _rename_name_in_importfrom(sc, ref.node, old_name, new_name)
        renames[old_name.rsplit('.', 1)[-1]] = new_name.rsplit('.', 1)[-1]
        already_changed.append(ref.node)
        has_changed = True

  for rename_old, rename_new in six.iteritems(renames):
    _rename_reads(sc, t, rename_old, rename_new)
  return has_changed


def _rename_name_in_importfrom(sc, node, old_name, new_name):
  if old_name == new_name:
    return False

  module_parts = node.module.split('.')
  old_parts = old_name.split('.')
  new_parts = new_name.split('.')

  # If just the module is changing, rename it
  if module_parts[:len(old_parts)] == old_parts:
    node.module = '.'.join(new_parts + module_parts[len(old_parts):])
    return True
    
  # Find the alias node to be changed
  for alias_to_change in node.names:
    if alias_to_change.name == old_parts[-1]:
      break
  else:
    return False

  alias_to_change.name = new_parts[-1]

  # Split the import if the package has changed
  if module_parts != new_parts[:-1]:
    if len(node.names) > 1:
      new_import = import_utils.split_import(sc, node, alias_to_change)
      new_import.module = '.'.join(new_parts[:-1])
    else:
      node.module = '.'.join(new_parts[:-1])

  return True


def _rename_reads(sc, t, old_name, new_name):
  """Updates all locations in the module where the given name is read.

  Arguments:
    sc: (scope.Scope) Scope to work in. This should be the scope of `t`.
    t: (ast.AST) The AST to perform updates in.
    old_name: (string) Dotted name to update.
    new_name: (string) Dotted name to replace it with.

  Returns:
    True if any changes were made, False otherwise.
  """
  name_parts = old_name.split('.')
  try:
    name = sc.names[name_parts[0]]
    for part in name_parts[1:]:
      name = name.attrs[part]
  except KeyError:
    return False

  has_changed = False
  for ref_node in name.reads:
    if isinstance(ref_node, (ast.Name, ast.Attribute)):
      ast_utils.replace_child(sc.parent(ref_node), ref_node,
                              ast.parse(new_name).body[0].value)
      has_changed = True

  return has_changed



# coding=utf-8
"""Helpers for working with fstrings (python3.6+)."""
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import ast

_FSTRING_VAL_PLACEHOLDER = '__pasta_fstring_val_{index}__'


def get_formatted_values(joined_str):
  """Get all FormattedValues from a JoinedStr, in order."""
  return [v for v in joined_str.values if isinstance(v, ast.FormattedValue)]


def placeholder(val_index):
  """Get the placeholder token for a FormattedValue in an fstring."""
  return _FSTRING_VAL_PLACEHOLDER.format(index=val_index)


def perform_replacements(fstr, values):
  """Replace placeholders in an fstring with subexpressions."""
  for i, value in enumerate(values):
    fstr = fstr.replace(_wrap(placeholder(i)), _wrap(value))
  return fstr


def _wrap(s):
  return '{%s}' % s



# coding=utf-8
"""Operations for storing and retrieving formatting info on ast nodes."""
# Copyright 2017 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

PASTA_DICT = '__pasta__'


def get(node, name, default=None):
  try:
    return _formatting_dict(node).get(name, default)
  except AttributeError:
    return default


def set(node, name, value):
  if not hasattr(node, PASTA_DICT):
    try:
      setattr(node, PASTA_DICT, {})
    except AttributeError:
      pass
  _formatting_dict(node)[name] = value


def append(node, name, value):
  set(node, name, get(node, name, '') + value)


def prepend(node, name, value):
  set(node, name, value + get(node, name, ''))


def _formatting_dict(node):
  return getattr(node, PASTA_DICT)



# coding=utf-8
"""Token generator for analyzing source code in logical units.

This module contains the TokenGenerator used for annotating a parsed syntax tree
with source code formatting.
"""
# Copyright 2017 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import ast
import collections
import contextlib
import itertools
import tokenize
from six import StringIO

from pasta.base import formatting as fmt
from pasta.base import fstring_utils

# Alias for extracting token names
TOKENS = tokenize
Token = collections.namedtuple('Token', ('type', 'src', 'start', 'end', 'line'))
FORMATTING_TOKENS = (TOKENS.INDENT, TOKENS.DEDENT, TOKENS.NL, TOKENS.NEWLINE,
                     TOKENS.COMMENT)


class TokenGenerator(object):
  """Helper for sequentially parsing Python source code, token by token.

  Holds internal state during parsing, including:
  _tokens: List of tokens in the source code, as parsed by `tokenize` module.
  _parens: Stack of open parenthesis at the current point in parsing.
  _hints: Number of open parentheses, brackets, etc. at the current point.
  _scope_stack: Stack containing tuples of nodes where the last parenthesis that
    was open is related to one of the nodes on the top of the stack.
  _lines: Full lines of the source code.
  _i: Index of the last token that was parsed. Initially -1.
  _loc: (lineno, column_offset) pair of the position in the source that has been
     parsed to. This should be either the start or end of the token at index _i.

  Arguments:
    ignore_error_tokens: If True, will ignore error tokens. Otherwise, an error
      token will cause an exception. This is useful when the source being parsed
      contains invalid syntax, e.g. if it is in an fstring context.
  """

  def __init__(self, source, ignore_error_token=False):
    self.lines = source.splitlines(True)
    self._tokens = list(_generate_tokens(source, ignore_error_token))
    self._parens = []
    self._hints = 0
    self._scope_stack = []
    self._len = len(self._tokens)
    self._i = -1
    self._loc = self.loc_begin()

  def chars_consumed(self):
    return len(self._space_between((1, 0), self._tokens[self._i].end))

  def loc_begin(self):
    """Get the start column of the current location parsed to."""
    if self._i < 0:
      return (1, 0)
    return self._tokens[self._i].start

  def loc_end(self):
    """Get the end column of the current location parsed to."""
    if self._i < 0:
      return (1, 0)
    return self._tokens[self._i].end

  def peek(self):
    """Get the next token without advancing."""
    if self._i + 1 >= self._len:
      return None
    return self._tokens[self._i + 1]

  def peek_non_whitespace(self):
    """Get the next non-whitespace token without advancing."""
    return self.peek_conditional(lambda t: t.type not in FORMATTING_TOKENS)

  def peek_conditional(self, condition):
    """Get the next token of the given type without advancing."""
    return next((t for t in self._tokens[self._i + 1:] if condition(t)), None)

  def next(self, advance=True):
    """Consume the next token and optionally advance the current location."""
    self._i += 1
    if self._i >= self._len:
      return None
    if advance:
      self._loc = self._tokens[self._i].end
    return self._tokens[self._i]

  def rewind(self, amount=1):
    """Rewind the token iterator."""
    self._i -= amount

  def whitespace(self, max_lines=None, comment=False):
    """Parses whitespace from the current _loc to the next non-whitespace.

    Arguments:
      max_lines: (optional int) Maximum number of lines to consider as part of
        the whitespace. Valid values are None, 0 and 1.
      comment: (boolean) If True, look for a trailing comment even when not in
        a parenthesized scope.

    Pre-condition:
      `_loc' represents the point before which everything has been parsed and
      after which nothing has been parsed.
    Post-condition:
      `_loc' is exactly at the character that was parsed to.
    """
    next_token = self.peek()
    if not comment and next_token and next_token.type == TOKENS.COMMENT:
      return ''
    def predicate(token):
      return (token.type in (TOKENS.INDENT, TOKENS.DEDENT) or
              token.type == TOKENS.COMMENT and (comment or self._hints) or
              token.type == TOKENS.ERRORTOKEN and token.src == ' ' or
              max_lines is None and token.type in (TOKENS.NL, TOKENS.NEWLINE))
    whitespace = list(self.takewhile(predicate, advance=False))
    next_token = self.peek()

    result = ''
    for tok in itertools.chain(whitespace,
                               ((next_token,) if next_token else ())):
      result += self._space_between(self._loc, tok.start)
      if tok != next_token:
        result += tok.src
        self._loc = tok.end
      else:
        self._loc = tok.start

    # Eat a single newline character
    if ((max_lines is None or max_lines > 0) and
        next_token and next_token.type in (TOKENS.NL, TOKENS.NEWLINE)):
      result += self.next().src

    return result

  def block_whitespace(self, indent_level):
    """Parses whitespace from the current _loc to the end of the block."""
    # Get the normal suffix lines, but don't advance the token index unless
    # there is no indentation to account for
    start_i = self._i
    full_whitespace = self.whitespace(comment=True)
    if not indent_level:
      return full_whitespace
    self._i = start_i

    # Trim the full whitespace into only lines that match the indentation level
    lines = full_whitespace.splitlines(True)
    try:
      last_line_idx = next(i for i, line in reversed(list(enumerate(lines)))
                           if line.startswith(indent_level + '#'))
    except StopIteration:
      # No comment lines at the end of this block
      self._loc = self._tokens[self._i].end
      return ''
    lines = lines[:last_line_idx + 1]

    # Advance the current location to the last token in the lines we've read
    end_line = self._tokens[self._i].end[0] + 1 + len(lines)
    list(self.takewhile(lambda tok: tok.start[0] < end_line))
    self._loc = self._tokens[self._i].end
    return ''.join(lines)

  def dots(self, num_dots):
    """Parse a number of dots.
    
    This is to work around an oddity in python3's tokenizer, which treats three
    `.` tokens next to each other in a FromImport's level as an ellipsis. This
    parses until the expected number of dots have been seen.
    """
    result = ''
    dots_seen = 0
    prev_loc = self._loc
    while dots_seen < num_dots:
      tok = self.next()
      assert tok.src in ('.', '...')
      result += self._space_between(prev_loc, tok.start) + tok.src
      dots_seen += tok.src.count('.')
      prev_loc = self._loc
    return result

  def open_scope(self, node, single_paren=False):
    """Open a parenthesized scope on the given node."""
    result = ''
    parens = []
    start_i = self._i
    start_loc = prev_loc = self._loc

    # Eat whitespace or '(' tokens one at a time
    for tok in self.takewhile(
        lambda t: t.type in FORMATTING_TOKENS or t.src == '('):
      # Stores all the code up to and including this token
      result += self._space_between(prev_loc, tok.start)

      if tok.src == '(' and single_paren and parens:
        self.rewind()
        self._loc = tok.start
        break

      result += tok.src
      if tok.src == '(':
        # Start a new scope
        parens.append(result)
        result = ''
        start_i = self._i
        start_loc = self._loc
      prev_loc = self._loc

    if parens:
      # Add any additional whitespace on to the last open-paren
      next_tok = self.peek()
      parens[-1] += result + self._space_between(self._loc, next_tok.start)
      self._loc = next_tok.start
      # Add each paren onto the stack
      for paren in parens:
        self._parens.append(paren)
        self._scope_stack.append(_scope_helper(node))
    else:
      # No parens were encountered, then reset like this method did nothing
      self._i = start_i
      self._loc = start_loc

  def close_scope(self, node, prefix_attr='prefix', suffix_attr='suffix',
                  trailing_comma=False, single_paren=False):
    """Close a parenthesized scope on the given node, if one is open."""
    # Ensures the prefix + suffix are not None
    if fmt.get(node, prefix_attr) is None:
      fmt.set(node, prefix_attr, '')
    if fmt.get(node, suffix_attr) is None:
      fmt.set(node, suffix_attr, '')

    if not self._parens or node not in self._scope_stack[-1]:
      return
    symbols = {')'}
    if trailing_comma:
      symbols.add(',')
    parsed_to_i = self._i
    parsed_to_loc = prev_loc = self._loc
    encountered_paren = False
    result = ''

    for tok in self.takewhile(
        lambda t: t.type in FORMATTING_TOKENS or t.src in symbols):
      # Consume all space up to this token
      result += self._space_between(prev_loc, tok.start)
      if tok.src == ')' and single_paren and encountered_paren:
        self.rewind()
        parsed_to_i = self._i
        parsed_to_loc = tok.start
        fmt.append(node, suffix_attr, result)
        break

      # Consume the token itself
      result += tok.src

      if tok.src == ')':
        # Close out the open scope
        encountered_paren = True
        self._scope_stack.pop()
        fmt.prepend(node, prefix_attr, self._parens.pop())
        fmt.append(node, suffix_attr, result)
        result = ''
        parsed_to_i = self._i
        parsed_to_loc = tok.end
        if not self._parens or node not in self._scope_stack[-1]:
          break
      prev_loc = tok.end

    # Reset back to the last place where we parsed anything
    self._i = parsed_to_i
    self._loc = parsed_to_loc

  def hint_open(self):
    """Indicates opening a group of parentheses or brackets."""
    self._hints += 1

  def hint_closed(self):
    """Indicates closing a group of parentheses or brackets."""
    self._hints -= 1
    if self._hints < 0:
      raise ValueError('Hint value negative')

  @contextlib.contextmanager
  def scope(self, node, attr=None, trailing_comma=False):
    """Context manager to handle a parenthesized scope."""
    self.open_scope(node, single_paren=(attr is not None))
    yield
    if attr:
      self.close_scope(node, prefix_attr=attr + '_prefix',
                       suffix_attr=attr + '_suffix',
                       trailing_comma=trailing_comma,
                       single_paren=True)
    else:
      self.close_scope(node, trailing_comma=trailing_comma)

  def is_in_scope(self):
    """Return True iff there is a scope open."""
    return self._parens or self._hints

  def str(self):
    """Parse a full string literal from the input."""
    def predicate(token):
      return (token.type in (TOKENS.STRING, TOKENS.COMMENT) or
              self.is_in_scope() and token.type in (TOKENS.NL, TOKENS.NEWLINE))

    return self.eat_tokens(predicate)

  def eat_tokens(self, predicate):
    """Parse input from tokens while a given condition is met."""
    content = ''
    prev_loc = self._loc
    tok = None
    for tok in self.takewhile(predicate, advance=False):
      content += self._space_between(prev_loc, tok.start)
      content += tok.src
      prev_loc = tok.end

    if tok:
      self._loc = tok.end
    return content

  def fstr(self):
    """Parses an fstring, including subexpressions.

    Returns:
      A generator function which, when repeatedly reads a chunk of the fstring
      up until the next subexpression and yields that chunk, plus a new token
      generator to use to parse the subexpression. The subexpressions in the
      original fstring data are replaced by placeholders to make it possible to
      fill them in with new values, if desired.
    """
    def fstr_parser():
      # Reads the whole fstring as a string, then parses it char by char
      if self.peek_non_whitespace().type == TOKENS.STRING:
        # Normal fstrings are one ore more STRING tokens, maybe mixed with
        # spaces, e.g.: f"Hello, {name}"
        str_content = self.str()
      else:
        # Format specifiers in fstrings are also JoinedStr nodes, but these are
        # arbitrary expressions, e.g. in: f"{value:{width}.{precision}}", the
        # format specifier is an fstring: "{width}.{precision}" but these are
        # not STRING tokens.
        def fstr_eater(tok):
          if tok.type == TOKENS.OP and tok.src == '}':
            if fstr_eater.level <= 0:
              return False
            fstr_eater.level -= 1
          if tok.type == TOKENS.OP and tok.src == '{':
            fstr_eater.level += 1
          return True
        fstr_eater.level = 0
        str_content = self.eat_tokens(fstr_eater)

      indexed_chars = enumerate(str_content)
      val_idx = 0
      i = -1
      result = ''
      while i < len(str_content) - 1:
        i, c = next(indexed_chars)
        result += c

        # When an open bracket is encountered, start parsing a subexpression
        if c == '{':
          # First check if this is part of an escape sequence
          # (f"{{" is used to escape a bracket literal)
          nexti, nextc = next(indexed_chars)
          if nextc == '{':
            result += c
            continue
          indexed_chars = itertools.chain([(nexti, nextc)], indexed_chars)

          # Add a placeholder onto the result
          result += fstring_utils.placeholder(val_idx) + '}'
          val_idx += 1

          # Yield a new token generator to parse the subexpression only
          tg = TokenGenerator(str_content[i+1:], ignore_error_token=True)
          yield (result, tg)
          result = ''

          # Skip the number of characters consumed by the subexpression
          for tg_i in range(tg.chars_consumed()):
            i, c = next(indexed_chars)

          # Eat up to and including the close bracket
          i, c = next(indexed_chars)
          while c != '}':
            i, c = next(indexed_chars)
      # Yield the rest of the fstring, when done
      yield (result, None)
    return fstr_parser

  def _space_between(self, start_loc, end_loc):
    """Parse the space between a location and the next token"""
    if start_loc > end_loc:
      raise ValueError('start_loc > end_loc', start_loc, end_loc)
    if start_loc[0] > len(self.lines):
      return ''

    prev_row, prev_col = start_loc
    end_row, end_col = end_loc
    if prev_row == end_row:
      return self.lines[prev_row - 1][prev_col:end_col]

    return ''.join(itertools.chain(
        (self.lines[prev_row - 1][prev_col:],),
        self.lines[prev_row:end_row - 1],
        (self.lines[end_row - 1][:end_col],) if end_col > 0 else '',
    ))

  def next_name(self):
    """Parse the next name token."""
    last_i = self._i
    def predicate(token):
      return token.type != TOKENS.NAME

    unused_tokens = list(self.takewhile(predicate, advance=False))
    result = self.next(advance=False)
    self._i = last_i
    return result

  def next_of_type(self, token_type):
    """Parse a token of the given type and return it."""
    token = self.next()
    if token.type != token_type:
      raise ValueError("Expected %r but found %r\nline %d: %s" % (
          tokenize.tok_name[token_type], token.src, token.start[0],
          self.lines[token.start[0] - 1]))
    return token

  def takewhile(self, condition, advance=True):
    """Parse tokens as long as a condition holds on the next token."""
    prev_loc = self._loc
    token = self.next(advance=advance)
    while token is not None and condition(token):
      yield token
      prev_loc = self._loc
      token = self.next(advance=advance)
    self.rewind()
    self._loc = prev_loc


def _scope_helper(node):
  """Get the closure of nodes that could begin a scope at this point.

  For instance, when encountering a `(` when parsing a BinOp node, this could
  indicate that the BinOp itself is parenthesized OR that the BinOp's left node
  could be parenthesized.

  E.g.: (a + b * c)   or   (a + b) * c   or   (a) + b * c
        ^                  ^                  ^

  Arguments:
    node: (ast.AST) Node encountered when opening a scope.

  Returns:
    A closure of nodes which that scope might apply to.
  """
  if isinstance(node, ast.Attribute):
    return (node,) + _scope_helper(node.value)
  if isinstance(node, ast.Subscript):
    return (node,) + _scope_helper(node.value)
  if isinstance(node, ast.Assign):
    return (node,) + _scope_helper(node.targets[0])
  if isinstance(node, ast.AugAssign):
    return (node,) + _scope_helper(node.target)
  if isinstance(node, ast.Expr):
    return (node,) + _scope_helper(node.value)
  if isinstance(node, ast.Compare):
    return (node,) + _scope_helper(node.left)
  if isinstance(node, ast.BoolOp):
    return (node,) + _scope_helper(node.values[0])
  if isinstance(node, ast.BinOp):
    return (node,) + _scope_helper(node.left)
  if isinstance(node, ast.Tuple) and node.elts:
    return (node,) + _scope_helper(node.elts[0])
  if isinstance(node, ast.Call):
    return (node,) + _scope_helper(node.func)
  if isinstance(node, ast.GeneratorExp):
    return (node,) + _scope_helper(node.elt)
  if isinstance(node, ast.IfExp):
    return (node,) + _scope_helper(node.body)
  return (node,)
   

def _generate_tokens(source, ignore_error_token=False):
  token_generator = tokenize.generate_tokens(StringIO(source).readline)
  try:
    for tok in token_generator:
      yield Token(*tok) 
  except tokenize.TokenError:
    if not ignore_error_token:
      raise



# coding=utf-8
"""Functions for dealing with import statements."""
# Copyright 2017 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import ast
import copy
import logging

from pasta.augment import errors
from pasta.base import ast_utils
from pasta.base import scope


def add_import(tree, name_to_import, asname=None, from_import=True, merge_from_imports=True):
  """Adds an import to the module.
  
  This function will try to ensure not to create duplicate imports. If name_to_import is
  already imported, it will return the existing import. This is true even if asname is set
  (asname will be ignored, and the existing name will be returned).
  
  If the import would create a name that already exists in the scope given by tree, this
  function will "import as", and append "_x" to the asname where x is the smallest positive
  integer generating a unique name.

  Arguments:
    tree: (ast.Module) Module AST to modify.
    name_to_import: (string) The absolute name to import.
    asname: (string) The alias for the import ("import name_to_import as asname")
    from_import: (boolean) If True, import the name using an ImportFrom node.
    merge_from_imports: (boolean) If True, merge a newly inserted ImportFrom
      node into an existing ImportFrom node, if applicable.

  Returns:
    The name (as a string) that can be used to reference the imported name. This
      can be the fully-qualified name, the basename, or an alias name.
  """
  sc = scope.analyze(tree)

  # Don't add anything if it's already imported
  if name_to_import in sc.external_references:
    existing_ref = next((ref for ref in sc.external_references[name_to_import]
                         if ref.name_ref is not None), None)
    if existing_ref:
      return existing_ref.name_ref.id

  import_node = None
  added_name = None
  
  def make_safe_alias_node(alias_name, asname):
    # Try to avoid name conflicts
    new_alias = ast.alias(name=alias_name, asname=asname)
    imported_name = asname or alias_name
    counter = 0
    while imported_name in sc.names:
      counter += 1
      imported_name = new_alias.asname = '%s_%d' % (asname or alias_name, 
                                                    counter)
    return new_alias
        
  # Add an ImportFrom node if requested and possible
  if from_import and '.' in name_to_import:
    from_module, alias_name = name_to_import.rsplit('.', 1)

    new_alias = make_safe_alias_node(alias_name, asname)
    
    if merge_from_imports:
      # Try to add to an existing ImportFrom from the same module
      existing_from_import = next(
          (node for node in tree.body if isinstance(node, ast.ImportFrom)
           and node.module == from_module and node.level == 0), None)
      if existing_from_import:
        existing_from_import.names.append(new_alias)
        return new_alias.asname or new_alias.name

    # Create a new node for this import
    import_node = ast.ImportFrom(module=from_module, names=[new_alias], level=0)

  # If not already created as an ImportFrom, create a normal Import node
  if not import_node:
    new_alias = make_safe_alias_node(alias_name=name_to_import, asname=asname)
    import_node = ast.Import(
        names=[new_alias])

  # Insert the node at the top of the module and return the name in scope
  tree.body.insert(1 if ast_utils.has_docstring(tree) else 0, import_node)
  return new_alias.asname or new_alias.name


def split_import(sc, node, alias_to_remove):
  """Split an import node by moving the given imported alias into a new import.

  Arguments:
    sc: (scope.Scope) Scope computed on whole tree of the code being modified.
    node: (ast.Import|ast.ImportFrom) An import node to split.
    alias_to_remove: (ast.alias) The import alias node to remove. This must be a
      child of the given `node` argument.

  Raises:
    errors.InvalidAstError: if `node` is not appropriately contained in the tree
      represented by the scope `sc`.
  """
  parent = sc.parent(node)
  parent_list = None
  for a in ('body', 'orelse', 'finalbody'):
    if hasattr(parent, a) and node in getattr(parent, a):
      parent_list = getattr(parent, a)
      break
  else:
    raise errors.InvalidAstError('Unable to find list containing import %r on '
                                 'parent node %r' % (node, parent))

  idx = parent_list.index(node)
  new_import = copy.deepcopy(node)
  new_import.names = [alias_to_remove]
  node.names.remove(alias_to_remove)

  parent_list.insert(idx + 1, new_import)
  return new_import


def get_unused_import_aliases(tree, sc=None):
  """Get the import aliases that aren't used.

  Arguments:
    tree: (ast.AST) An ast to find imports in.
    sc: A scope.Scope representing tree (generated from scratch if not
    provided).

  Returns:
    A list of ast.alias representing imported aliases that aren't referenced in
    the given tree.
  """
  if sc is None:
    sc = scope.analyze(tree)
  unused_aliases = set()
  for node in ast.walk(tree):
    if isinstance(node, ast.alias):
      str_name = node.asname if node.asname is not None else node.name
      if str_name in sc.names:
        name = sc.names[str_name]
        if not name.reads:
          unused_aliases.add(node)
      else:
        # This happens because of https://github.com/google/pasta/issues/32
        logging.warning('Imported name %s not found in scope (perhaps it\'s '
                        'imported dynamically)', str_name)

  return unused_aliases


def remove_import_alias_node(sc, node):
  """Remove an alias and if applicable remove their entire import.

  Arguments:
    sc: (scope.Scope) Scope computed on whole tree of the code being modified.
    node: (ast.Import|ast.ImportFrom|ast.alias) The node to remove.
  """
  import_node = sc.parent(node)
  if len(import_node.names) == 1:
    import_parent = sc.parent(import_node)
    ast_utils.remove_child(import_parent, import_node)
  else:
    ast_utils.remove_child(import_node, node)


def remove_duplicates(tree, sc=None):
  """Remove duplicate imports, where it is safe to do so.

  This does NOT remove imports that create new aliases

  Arguments:
    tree: (ast.AST) An ast to modify imports in.
    sc: A scope.Scope representing tree (generated from scratch if not
    provided).

  Returns:
    Whether any changes were made.
  """
  if sc is None:
    sc = scope.analyze(tree)

  modified = False
  seen_names = set()
  for node in tree.body:
    if isinstance(node, (ast.Import, ast.ImportFrom)):
      for alias in list(node.names):
        import_node = sc.parent(alias)
        if isinstance(import_node, ast.Import):
          full_name = alias.name
        elif import_node.module:
          full_name = '%s%s.%s' % ('.' * import_node.level,
                                   import_node.module, alias.name)
        else:
          full_name = '%s%s' % ('.' * import_node.level, alias.name)
        full_name += ':' + (alias.asname or alias.name)
        if full_name in seen_names:
          remove_import_alias_node(sc, alias)
          modified = True
        else:
          seen_names.add(full_name)
  return modified



# coding=utf-8
"""Tests for google3.third_party.py.pasta.base.test_utils."""
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import ast
import unittest

from pasta.base import test_utils


class CheckAstEqualityTest(test_utils.TestCase):

  def test_empty(self):
    src = ""
    t = ast.parse(src)
    self.checkAstsEqual(t, t)

  def test_one_global(self):
    src = "X = 1\n"
    t = ast.parse(src)
    self.checkAstsEqual(t, t)

  def test_two_globals(self):
    src = "X = 1\nY = 2\n"
    t = ast.parse(src)
    self.checkAstsEqual(t, t)

  def test_different_number_of_nodes(self):
    src1 = "X = 1\ndef Foo():\n  return None\n"
    src2 = src1 + "Y = 2\n"
    t1 = ast.parse(src1)
    t2 = ast.parse(src2)
    with self.assertRaises(AssertionError):
      self.checkAstsEqual(t1, t2)

  def test_simple_function_def(self):
    code = ("def foo(x):\n"
            "  return x + 1\n")
    t = ast.parse(code)
    self.checkAstsEqual(t, t)


def suite():
  result = unittest.TestSuite()
  result.addTests(unittest.makeSuite(CheckAstEqualityTest))
  return result


if __name__ == '__main__':
  unittest.main()



# coding=utf-8
"""Perform static analysis on python syntax trees."""
# Copyright 2017 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import ast
import collections
import six

# TODO: Support relative imports

# Represents a reference to something external to the module.
# Fields:
#   name: (string) The full dotted name being referenced.
#   node: (ast.AST) The AST node where the reference is defined.
#   name_ref: (Name) The name object that refers to the imported name, if
#     applicable. This may not be the same id if the import is aliased.
ExternalReference = collections.namedtuple('ExternalReference',
                                           ('name', 'node', 'name_ref'))


class ScopeVisitor(ast.NodeVisitor):

  def __init__(self):
    super(ScopeVisitor, self).__init__()
    self._parent = None
    self.root_scope = self.scope = RootScope(None)

  def visit(self, node):
    if node is None:
      return
    if self.root_scope.node is None:
      self.root_scope.node = node
    self.root_scope.set_parent(node, self._parent)
    tmp = self._parent
    self._parent = node
    super(ScopeVisitor, self).visit(node)
    self._parent = tmp

  def visit_in_order(self, node, *attrs):
    for attr in attrs:
      val = getattr(node, attr, None)
      if val is None:
        continue
      if isinstance(val, list):
        for item in val:
          self.visit(item)
      elif isinstance(val, ast.AST):
        self.visit(val)

  def visit_Import(self, node):
    for alias in node.names:
      name_parts = alias.name.split('.')

      if not alias.asname:
        # If not aliased, define the top-level module of the import
        cur_name = self.scope.define_name(name_parts[0], alias)
        self.root_scope.add_external_reference(name_parts[0], alias,
                                               name_ref=cur_name)

        # Define names of sub-modules imported
        partial_name = name_parts[0]
        for part in name_parts[1:]:
          partial_name += '.' + part
          cur_name = cur_name.lookup_name(part)
          cur_name.define(alias)
          self.root_scope.add_external_reference(partial_name, alias,
                                                 name_ref=cur_name)

      else:
        # If the imported name is aliased, define that name only
        name = self.scope.define_name(alias.asname, alias)

        # Define names of sub-modules imported
        for i in range(1, len(name_parts)):
          self.root_scope.add_external_reference('.'.join(name_parts[:i]),
                                                 alias)
        self.root_scope.add_external_reference(alias.name, alias, name_ref=name)

    self.generic_visit(node)

  def visit_ImportFrom(self, node):
    if node.module:
      name_parts = node.module.split('.')
      for i in range(1, len(name_parts) + 1):
        self.root_scope.add_external_reference('.'.join(name_parts[:i]), node)
    for alias in node.names:
      name = self.scope.define_name(alias.asname or alias.name, alias)
      if node.module:
        self.root_scope.add_external_reference(
            '.'.join((node.module, alias.name)), alias, name_ref=name)
      # TODO: else? relative imports
    self.generic_visit(node)

  def visit_Name(self, node):
    if isinstance(node.ctx, (ast.Store, ast.Param)):
      self.scope.define_name(node.id, node)
    elif isinstance(node.ctx, ast.Load):
      self.scope.lookup_name(node.id).add_reference(node)
      self.root_scope.set_name_for_node(node, self.scope.lookup_name(node.id))
    self.generic_visit(node)

  def visit_FunctionDef(self, node):
    # Visit decorator list first to avoid declarations in args
    self.visit_in_order(node, 'decorator_list')
    if isinstance(self.root_scope.parent(node), ast.ClassDef):
      pass # TODO: Support referencing methods by "self" where possible
    else:
      self.scope.define_name(node.name, node)
    try:
      self.scope = self.scope.create_scope(node)
      self.visit_in_order(node, 'args', 'returns', 'body')
    finally:
      self.scope = self.scope.parent_scope

  def visit_arguments(self, node):
    self.visit_in_order(node, 'defaults', 'args')
    if six.PY2:
      # In python 2.x, these names are not Name nodes. Define them explicitly
      # to be able to find references in the function body.
      for arg_attr_name in ('vararg', 'kwarg'):
        arg_name = getattr(node, arg_attr_name, None)
        if arg_name is not None:
          self.scope.define_name(arg_name, node)
    else:
      # Visit defaults first to avoid declarations in args
      self.visit_in_order(node, 'vararg', 'kwarg')

  def visit_arg(self, node):
    self.scope.define_name(node.arg, node)
    self.generic_visit(node)

  def visit_ClassDef(self, node):
    self.visit_in_order(node, 'decorator_list', 'bases')
    self.scope.define_name(node.name, node)
    try:
      self.scope = self.scope.create_scope(node)
      self.visit_in_order(node, 'body')
    finally:
      self.scope = self.scope.parent_scope

  def visit_Attribute(self, node):
    self.generic_visit(node)
    node_value_name = self.root_scope.get_name_for_node(node.value)
    if node_value_name:
      node_name = node_value_name.lookup_name(node.attr)
      self.root_scope.set_name_for_node(node, node_name)
      node_name.add_reference(node)


class Scope(object):

  def __init__(self, parent_scope, node):
    self.parent_scope = parent_scope
    self.names = {}
    self.node = node

  def define_name(self, name, node):
    try:
      name_obj = self.names[name]
    except KeyError:
      name_obj = self.names[name] = Name(name)
    name_obj.define(node)
    return name_obj

  def lookup_name(self, name):
    try:
      return self.names[name]
    except KeyError:
      pass
    if self.parent_scope is None:
      name_obj = self.names[name] = Name(name)
      return name_obj
    return self.parent_scope.lookup_name(name)

  def get_root_scope(self):
    return self.parent_scope.get_root_scope()

  def lookup_scope(self, node):
    return self.get_root_scope().lookup_scope(node)

  def create_scope(self, node):
    subscope = Scope(self, node)
    self.get_root_scope()._set_scope_for_node(node, subscope)
    return subscope


class RootScope(Scope):

  def __init__(self, node):
    super(RootScope, self).__init__(None, node)
    self.external_references = {}
    self._parents = {}
    self._nodes_to_names = {}
    self._node_scopes = {}

  def add_external_reference(self, name, node, name_ref=None):
    ref = ExternalReference(name=name, node=node, name_ref=name_ref)
    if name in self.external_references:
      self.external_references[name].append(ref)
    else:
      self.external_references[name] = [ref]

  def get_root_scope(self):
    return self

  def parent(self, node):
    return self._parents.get(node, None)

  def set_parent(self, node, parent):
    self._parents[node] = parent
    if parent is None:
      self._node_scopes[node] = self

  def get_name_for_node(self, node):
    return self._nodes_to_names.get(node, None)

  def set_name_for_node(self, node, name):
    self._nodes_to_names[node] = name

  def lookup_scope(self, node):
    while node:
      try:
        return self._node_scopes[node]
      except KeyError:
        node = self.parent(node)
    return None

  def _set_scope_for_node(self, node, node_scope):
    self._node_scopes[node] = node_scope


# Should probably also have a scope?
class Name(object):

  def __init__(self, id):
    self.id = id
    self.definition = None
    self.reads = []
    self.attrs = {}

  def add_reference(self, node):
    self.reads.append(node)

  def define(self, node):
    if self.definition:
      self.reads.append(node)
    else:
      self.definition = node

  def lookup_name(self, name):
    try:
      return self.attrs[name]
    except KeyError:
      name_obj = self.attrs[name] = Name('.'.join((self.id, name)))
      return name_obj


def analyze(tree):
  v = ScopeVisitor()
  v.visit(tree)
  return v.scope



"""Constants relevant to ast code."""

import ast

NODE_TYPE_TO_TOKENS = {
    ast.Add: ('+',),
    ast.And: ('and',),
    ast.BitAnd: ('&',),
    ast.BitOr: ('|',),
    ast.BitXor: ('^',),
    ast.Div: ('/',),
    ast.Eq: ('==',),
    ast.FloorDiv: ('//',),
    ast.Gt: ('>',),
    ast.GtE: ('>=',),
    ast.In: ('in',),
    ast.Invert: ('~',),
    ast.Is: ('is',),
    ast.IsNot: ('is', 'not',),
    ast.LShift: ('<<',),
    ast.Lt: ('<',),
    ast.LtE: ('<=',),
    ast.Mod: ('%',),
    ast.Mult: ('*',),
    ast.Not: ('not',),
    ast.NotEq: ('!=',),
    ast.NotIn: ('not', 'in',),
    ast.Or: ('or',),
    ast.Pow: ('**',),
    ast.RShift: ('>>',),
    ast.Sub: ('-',),
    ast.UAdd: ('+',),
    ast.USub: ('-',),
}


if hasattr(ast, 'MatMult'):
  NODE_TYPE_TO_TOKENS[ast.MatMult] = ('@',)



# coding=utf-8
"""Tests for ast_utils."""
# Copyright 2017 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pasta
from pasta.augment import errors
from pasta.base import ast_utils
from pasta.base import test_utils


class UtilsTest(test_utils.TestCase):

  def test_sanitize_source(self):
    coding_lines = (
        '# -*- coding: latin-1 -*-',
        '# -*- coding: iso-8859-15 -*-',
        '# vim: set fileencoding=ascii :',
        '# This Python file uses the following encoding: utf-8',
    )
    src_template = '{coding}\na = 123\n'
    sanitized_src = '# (removed coding)\na = 123\n'
    for line in coding_lines:
      src = src_template.format(coding=line)

      # Replaced on lines 1 and 2
      self.assertEqual(sanitized_src, ast_utils.sanitize_source(src))
      src_prefix = '"""Docstring."""\n'
      self.assertEqual(src_prefix + sanitized_src,
                       ast_utils.sanitize_source(src_prefix + src))

      # Unchanged on line 3
      src_prefix = '"""Docstring."""\n# line 2\n'
      self.assertEqual(src_prefix + src,
                       ast_utils.sanitize_source(src_prefix + src))


class AlterChildTest(test_utils.TestCase):

  def testRemoveChildMethod(self):
    src = """\
class C():
  def f(x):
    return x + 2
  def g(x):
    return x + 3
"""
    tree = pasta.parse(src)
    class_node = tree.body[0]
    meth1_node = class_node.body[0]

    ast_utils.remove_child(class_node, meth1_node)

    result = pasta.dump(tree)
    expected = """\
class C():
  def g(x):
    return x + 3
"""
    self.assertEqual(result, expected)

  def testRemoveAlias(self):
    src = "from a import b, c"
    tree = pasta.parse(src)
    import_node = tree.body[0]
    alias1 = import_node.names[0]
    ast_utils.remove_child(import_node, alias1)

    self.assertEqual(pasta.dump(tree), "from a import c")

  def testRemoveFromBlock(self):
    src = """\
if a:
  print("foo!")
  x = 1
"""
    tree = pasta.parse(src)
    if_block = tree.body[0]
    print_stmt = if_block.body[0]
    ast_utils.remove_child(if_block, print_stmt)

    expected = """\
if a:
  x = 1
"""
    self.assertEqual(pasta.dump(tree), expected)

  def testReplaceChildInBody(self):
    src = 'def foo():\n  a = 0\n  a += 1 # replace this\n  return a\n'
    replace_with = pasta.parse('foo(a + 1)  # trailing comment\n').body[0]
    expected = 'def foo():\n  a = 0\n  foo(a + 1) # replace this\n  return a\n'
    t = pasta.parse(src)

    parent = t.body[0]
    node_to_replace = parent.body[1]
    ast_utils.replace_child(parent, node_to_replace, replace_with)

    self.assertEqual(expected, pasta.dump(t))

  def testReplaceChildInvalid(self):
    src = 'def foo():\n  return 1\nx = 1\n'
    replace_with = pasta.parse('bar()').body[0]
    t = pasta.parse(src)

    parent = t.body[0]
    node_to_replace = t.body[1]
    with self.assertRaises(errors.InvalidAstError):
      ast_utils.replace_child(parent, node_to_replace, replace_with)
